# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from ai4rag.search_space.src.models import FoundationModels

__all__ = [
    "get_system_message_text",
    "get_user_message_text",
    "get_context_template_text",
    "QUESTION_PLACEHOLDER",
    "REFERENCE_DOCUMENTS_PLACEHOLDER",
    "CONTEXT_TEXT_PLACEHOLDER",
    "MULTILINGUAL_SUPPORT_INSTRUCTION_PLACEHOLDER",
]


QUESTION_PLACEHOLDER = "question"
REFERENCE_DOCUMENTS_PLACEHOLDER = "reference_documents"
CONTEXT_TEXT_PLACEHOLDER = "document"
MULTILINGUAL_SUPPORT_INSTRUCTION_PLACEHOLDER = "multilingual_support"


# A mapping from model name into their corresponding prompt templates.
# The parameters for the prompt templates are QUESTION_PLACEHOLDER and REFERENCE_DOCUMENTS_PLACEHOLDER

_MULTILINGUAL_SUPPORT_ENABLED_PROMPT = (
    "Respond exclusively in the language of the question, "
    "regardless of any other language used in the provided context. "
    "Ensure that your entire response is in the same language as the question."
)


_MULTILINGUAL_SUPPORT_DISABLED_PROMPT = (
    "Respond exclusively in English, "
    "regardless of the language of the question or any other language used in the provided context. "
    "Ensure that your entire response is in English only."
)


_DEFAULT_SYSTEM_MESSAGE_TEXT = (
    "Please answer the question I provide in the Question section below, "
    "based solely on the information I provide in the Context section. "
    "If the question is unanswerable, please say you cannot answer."
)


_DEFAULT_USER_MESSAGE_TEXT = (
    f"\n\nContext:\n{{{REFERENCE_DOCUMENTS_PLACEHOLDER}}}:\n\n"
    f"Question: {{{QUESTION_PLACEHOLDER}}}. \n"
    "Again, please answer the question based on the context provided only. If the context is not related to "
    "the question, just say you cannot answer. "
    f"{{{MULTILINGUAL_SUPPORT_INSTRUCTION_PLACEHOLDER}}}"
)


_DEFAULT_GRANITE_SYSTEM_MESSAGE_TEXT = (
    "You are Granite Chat, an AI language model developed by IBM. "
    "You are a cautious assistant. You carefully follow instructions. "
    "You are helpful and harmless and you follow ethical guidelines and promote positive behaviour."
)


_DEFAULT_GRANITE_USER_MESSAGE_TEXT = (
    "You are an AI language model designed to function as a specialized Retrieval Augmented Generation (RAG) "
    "assistant. When generating responses, prioritize correctness, i.e., ensure that your response is grounded in "
    "context and user query. Always make sure that your response is relevant to the question. "
    "\n"
    "Answer Length: detailed"
    "\n"
    f"{{{REFERENCE_DOCUMENTS_PLACEHOLDER}}}"
    "\n"
    f"{{{MULTILINGUAL_SUPPORT_INSTRUCTION_PLACEHOLDER}}}"
    "\n"
    f"{{{QUESTION_PLACEHOLDER}}} "
    "\n"
    "\n"
)


_DEFAULT_LLAMA_SYSTEM_MESSAGE_TEXT = (
    "You are a helpful, respectful and honest assistant. Always answer as helpfully as possible, while being safe. "
    "Your answers should not include any harmful, unethical, racist, sexist, toxic, dangerous, or illegal content. "
    "Please ensure that your responses are socially unbiased and positive in nature.\n"
    "If a question does not make any sense, or is not factually coherent, explain why instead of answering "
    "something not correct. If you don’t know the answer to a question, please don’t share false information.\n"
)


_DEFAULT_LLAMA_USER_MESSAGE_TEXT = (
    f"{{{REFERENCE_DOCUMENTS_PLACEHOLDER}}}\n"
    f"[conversation]: {{{QUESTION_PLACEHOLDER}}}. Answer with no more than 150 words. If you cannot base your "
    "answer on the given document, please state that you do not have an answer. "
    f"{{{MULTILINGUAL_SUPPORT_INSTRUCTION_PLACEHOLDER}}}\n"
)


_DEFAULT_MISTRAL_SYSTEM_MESSAGE_TEXT = (
    "You are a helpful, respectful and honest assistant. "
    "Always answer as helpfully as possible, while being safe. "
    "Your answers should not include any harmful, unethical, racist, sexist, toxic, dangerous, or illegal content. "
    "Please ensure that your responses are socially unbiased and positive in nature.\n\n"
    "If a question does not make any sense, or is not factually coherent, explain why instead of answering "
    "something not correct. If you don't know the answer to a question, please don't share false information.\n\n"
)


_DEFAULT_MISTRAL_USER_MESSAGE_TEXT = (
    "Generate the next agent response by answering the question. You are provided several documents with titles. "
    "If the answer comes from different documents please mention all possibilities and use the titles of documents "
    "to separate between topics or domains. If you cannot base your answer on the given documents, "
    f"please state that you do not have an answer. "
    f"{{{REFERENCE_DOCUMENTS_PLACEHOLDER}}}\n\n"
    f"{{{MULTILINGUAL_SUPPORT_INSTRUCTION_PLACEHOLDER}}}\n\n"
    f"{{{QUESTION_PLACEHOLDER}}}"
)


_DEFAULT_OPENAI_SYSTEM_MESSAGE_TEXT = (
    "You are a AI language model designed to function as a specialized Retrieval Augmented Generation (RAG) assistant. "
    "When generating responses, prioritize correctness, i.e., ensure that your response is correct given the context "
    "and user query, and that it is grounded in the context. "
    "Furthermore, make sure that the response is supported by the given document or context. "
    "When the question cannot be answered using the context or document, output the following response: "
    "'I am sorry, I do not have the information you are looking for in my knowledge base.'. "
    "Always make sure that your response is relevant to the question. If an explanation is needed, "
    "first provide the explanation or reasoning, and then give the final answer.\nAnswer Length: concise.\n\n"
)


_DEFAULT_OPENAI_USER_MESSAGE_TEXT = (
    f"[Document]\n{{{REFERENCE_DOCUMENTS_PLACEHOLDER}}}\n[End]\n"
    f"{{{QUESTION_PLACEHOLDER}}}. \n"
    f"{{{MULTILINGUAL_SUPPORT_INSTRUCTION_PLACEHOLDER}}}"
)


_model_name_to_system_message_text = {
    FoundationModels.META_LLAMA_3_1_70B_INSTRUCT: _DEFAULT_LLAMA_SYSTEM_MESSAGE_TEXT,
    FoundationModels.META_LLAMA_3_1_8B_INSTRUCT: _DEFAULT_LLAMA_SYSTEM_MESSAGE_TEXT,
    FoundationModels.META_LLAMA_3_3_70B_INSTRUCT: _DEFAULT_LLAMA_SYSTEM_MESSAGE_TEXT,
    FoundationModels.META_LLAMA_4_MAVERICK_17B_128E_INSTRUCT_FP8: _DEFAULT_LLAMA_SYSTEM_MESSAGE_TEXT,
    FoundationModels.GRANITE_3_8B_INSTRUCT: _DEFAULT_GRANITE_SYSTEM_MESSAGE_TEXT,
    FoundationModels.GRANITE_3_3_8B_INSTRUCT: _DEFAULT_GRANITE_SYSTEM_MESSAGE_TEXT,
    FoundationModels.MISTRAL_SMALL_3_1_24B_INSTRUCT: _DEFAULT_MISTRAL_SYSTEM_MESSAGE_TEXT,
    FoundationModels.MISTRAL_MEDIUM_2505: _DEFAULT_MISTRAL_SYSTEM_MESSAGE_TEXT,
    FoundationModels.MISTRAL_MISTRAL_LARGE: _DEFAULT_MISTRAL_SYSTEM_MESSAGE_TEXT,
    FoundationModels.OPENAI_GPT_OSS_120B: _DEFAULT_OPENAI_SYSTEM_MESSAGE_TEXT,
}


_model_name_to_user_message_text = {
    FoundationModels.META_LLAMA_3_1_70B_INSTRUCT: _DEFAULT_LLAMA_USER_MESSAGE_TEXT,
    FoundationModels.META_LLAMA_3_1_8B_INSTRUCT: _DEFAULT_LLAMA_USER_MESSAGE_TEXT,
    FoundationModels.META_LLAMA_3_3_70B_INSTRUCT: _DEFAULT_LLAMA_USER_MESSAGE_TEXT,
    FoundationModels.META_LLAMA_4_MAVERICK_17B_128E_INSTRUCT_FP8: _DEFAULT_LLAMA_USER_MESSAGE_TEXT,
    FoundationModels.GRANITE_3_8B_INSTRUCT: _DEFAULT_GRANITE_USER_MESSAGE_TEXT,
    FoundationModels.GRANITE_3_3_8B_INSTRUCT: _DEFAULT_GRANITE_USER_MESSAGE_TEXT,
    FoundationModels.MISTRAL_SMALL_3_1_24B_INSTRUCT: _DEFAULT_MISTRAL_USER_MESSAGE_TEXT,
    FoundationModels.MISTRAL_MEDIUM_2505: _DEFAULT_MISTRAL_USER_MESSAGE_TEXT,
    FoundationModels.MISTRAL_MISTRAL_LARGE: _DEFAULT_MISTRAL_USER_MESSAGE_TEXT,
    FoundationModels.OPENAI_GPT_OSS_120B: _DEFAULT_OPENAI_USER_MESSAGE_TEXT,
}


# A mapping from model names into their corresponding context template texts. These templates describe how each
# retrieved context is to be wrapped, before being integrated into a full RAG prompt text.
# The parameter for the context template text is CONTEXT_TEXT_PLACEHOLDER
_DEFAULT_GRANITE_CONTEXT_TEMPLATE = f"[Document]\n{{{CONTEXT_TEXT_PLACEHOLDER}}}\n[End]"
_DEFAULT_LLAMA_CONTEXT_TEMPLATE = f"[document]: {{{CONTEXT_TEXT_PLACEHOLDER}}}\n"
_DEFAULT_CONTEXT_TEMPLATE = f"{{{CONTEXT_TEXT_PLACEHOLDER}}}"

_model_name_to_context_template_text = {
    FoundationModels.META_LLAMA_3_1_70B_INSTRUCT: _DEFAULT_LLAMA_CONTEXT_TEMPLATE,
    FoundationModels.META_LLAMA_3_1_8B_INSTRUCT: _DEFAULT_LLAMA_CONTEXT_TEMPLATE,
    FoundationModels.META_LLAMA_3_3_70B_INSTRUCT: _DEFAULT_LLAMA_CONTEXT_TEMPLATE,
    FoundationModels.META_LLAMA_4_MAVERICK_17B_128E_INSTRUCT_FP8: _DEFAULT_LLAMA_CONTEXT_TEMPLATE,
    FoundationModels.GRANITE_3_8B_INSTRUCT: _DEFAULT_GRANITE_CONTEXT_TEMPLATE,
    FoundationModels.GRANITE_3_3_8B_INSTRUCT: _DEFAULT_GRANITE_CONTEXT_TEMPLATE,
    FoundationModels.OPENAI_GPT_OSS_120B: _DEFAULT_CONTEXT_TEMPLATE,
}


def get_context_template_text(model_name: str) -> str:
    """
    Get a model-specific context template text.

    The context template text is a template with one placeholder: "context_text".
    This field should be populated before use within a RAG prompt.

    Parameters
    ----------
    model_name : str
        The name of the model for which we should return the context template text.

    Returns
    -------
    str
        The context template text str for the given model name.
    """
    context_template = _model_name_to_context_template_text.get(model_name, None)

    if not context_template:
        if "granite" in model_name:
            context_template = _DEFAULT_GRANITE_CONTEXT_TEMPLATE
        elif "llama" in model_name:
            context_template = _DEFAULT_LLAMA_CONTEXT_TEMPLATE
        else:
            context_template = _DEFAULT_CONTEXT_TEMPLATE

    return context_template


def get_system_message_text(model_name: str) -> str:
    """
    Get model-specific system prompt text.

    Parameters
    ----------
    model_name : str
        The name of the model for which we should return the system prompt text.

    Returns
    -------
    str
        The system prompt text str for the given model name.
    """

    system_message_text = _model_name_to_system_message_text.get(model_name, None)

    if not system_message_text:
        if "granite" in model_name:
            system_message_text = _DEFAULT_GRANITE_SYSTEM_MESSAGE_TEXT
        elif "llama" in model_name:
            system_message_text = _DEFAULT_LLAMA_SYSTEM_MESSAGE_TEXT
        elif "mistral" in model_name:
            system_message_text = _DEFAULT_MISTRAL_SYSTEM_MESSAGE_TEXT
        else:
            system_message_text = _DEFAULT_SYSTEM_MESSAGE_TEXT

    return system_message_text


def get_user_message_text(model_name: str, language_autodetect: bool = True) -> str:
    """
    Get a model-specific prompt text.

    The user message text is a template, with two markers for fields: "question" and "reference_documents".
    These fields should be filled appropriately before delivering the prompt to a model.

    Parameters
    ----------
    model_name : str
        The name of the model for which we should return the prompt text.

    language_autodetect : bool
        If True, language of the question will be automatically detected.

    Returns
    -------
    str
        The prompt text str matching the given model name.
    """
    user_message_text = _model_name_to_user_message_text.get(model_name, None)

    if not user_message_text:
        if "granite" in model_name:
            user_message_text = _DEFAULT_GRANITE_USER_MESSAGE_TEXT
        elif "llama" in model_name:
            user_message_text = _DEFAULT_LLAMA_USER_MESSAGE_TEXT
        elif "mistral" in model_name:
            user_message_text = _DEFAULT_MISTRAL_USER_MESSAGE_TEXT
        else:
            user_message_text = _DEFAULT_USER_MESSAGE_TEXT

    user_message_text = user_message_text.replace(
        "{multilingual_support}",
        _MULTILINGUAL_SUPPORT_ENABLED_PROMPT if language_autodetect else _MULTILINGUAL_SUPPORT_DISABLED_PROMPT,
    )

    return user_message_text



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Any, TypedDict

from langchain_core.documents import Document

from ai4rag import logger
from ai4rag.core.experiment.benchmark_data import BenchmarkData
from ai4rag.core.experiment.exception_handler import (
    EvaluationError,
    ExperimentExceptionHandler,
    GenerationError,
    IndexingError,
)
from ai4rag.core.experiment.utils import build_evaluation_data, query_rag
from ai4rag.evaluator import UnitxtEvaluator
from ai4rag.evaluator.base_evaluator import BaseEvaluator
from ai4rag.rag.chunking.langchain_chunker import LangChainChunker
from ai4rag.rag.embedding.base_model import BaseEmbeddingModel
from ai4rag.rag.foundation_models.base_model import BaseFoundationModel
from ai4rag.rag.retrieval.retriever import Retriever
from ai4rag.rag.template.simple_rag_template import SimpleRAG
from ai4rag.rag.vector_store.base_vector_store import BaseVectorStore
from ai4rag.rag.vector_store.chroma import ChromaVectorStore
from ai4rag.utils.constants import AI4RAGParamNames

__all__ = ["PreSelectorError", "ModelsPreSelector"]


class PreSelectorError(Exception):
    """Exception to be raised when critical issue occurs in the MPS."""


class MPSEvaluationResultsTyped(TypedDict):
    """Typing helper for evaluation results coming from ModelsPreSelector."""

    foundation_model: BaseFoundationModel
    embedding_model: BaseEmbeddingModel
    scores: dict
    question_scores: dict


# pylint: disable=too-many-instance-attributes
class ModelsPreSelector:
    """
    Class responsible for performing foundation and embeddings models preselection.
    Using sample of benchmark_data and sample of grounding documents
    ModelsPreSelector is able to evaluate which top models
    should be selected as the best promising ones for the further experiment steps.

    ModelsPreSelector performs RAG pattern evaluation for each
    foundation model and embedding model pair with pre-configured settings
    using data sample. It provides the best performing pairs of generation and
    embedding models, that are considered further during HPO.

    Parameters
    ----------
    metric : str
        Metric used in ranking the models.

    foundation_models : list[BaseFoundationModel]
        List of foundation models that should be considered in the selection.

    embedding_models : list[BaseEmbeddingModel]
        Embedding models to models pre-selection.

    documents : list[Document]
        Grounding documents that will be sampled to perform pre-selection.

    benchmark_data : BenchmarkData
        Sample of benchmark data used for the pre-selection.

    Attributes
    ----------
    evaluator : BaseEvaluator
        Instance responsible for RAG pattern's response evaluation.

    retrieval_params : dict
        Retrieval parameters for all MPS evaluations.

    chunking_params : dict
        Chunking parameters for all MPS evaluations.

    evaluation_results : list[MPSEvaluationResultsTyped]
        Dictionary holding results from evaluating each RAG Pattern.
        This may be overwritten by the user to avoid evaluation and
        pre-select models based on mean scores.

    DEFAULT_N_FOUNDATION_MODELS : int
        Number of foundation models to select in the process of MPS.

    DEFAULT_N_EMBEDDING_MODELS : int
        Number of embedding models to select in the process of MPS.
    """

    DEFAULT_N_FOUNDATION_MODELS = 3
    DEFAULT_N_EMBEDDING_MODELS = 2

    def __init__(
        self,
        metric: str,
        foundation_models: list[BaseFoundationModel],
        embedding_models: list[BaseEmbeddingModel],
        documents: list[Document],
        benchmark_data: BenchmarkData,
        **kwargs,
    ):
        self.benchmark_data = benchmark_data
        self.documents = documents
        self.foundation_models = foundation_models
        self.embedding_models = embedding_models
        self.metric = metric

        self.evaluator: BaseEvaluator = kwargs.get("evaluator", UnitxtEvaluator())
        self.retrieval_params = {
            "number_of_chunks": kwargs.get(AI4RAGParamNames.NUMBER_OF_CHUNKS, 3),
            "method": kwargs.get(AI4RAGParamNames.RETRIEVAL_METHOD, "simple"),
            "search_mode": kwargs.get(AI4RAGParamNames.SEARCH_MODE, "vector"),
        }
        self.chunking_params = {
            "chunk_size": kwargs.get(AI4RAGParamNames.CHUNK_SIZE, 512),
            "method": kwargs.get(AI4RAGParamNames.CHUNKING_METHOD, "recursive"),
            "chunk_overlap": kwargs.get(AI4RAGParamNames.CHUNK_OVERLAP, 128),
        }
        self.evaluation_results: list[MPSEvaluationResultsTyped] = []
        self._exception_handler = ExperimentExceptionHandler()

    def evaluate_patterns(self):
        """
        Evaluate RAG pattern per each foundation model provided. All settings
        of the patterns are configured and remain the same for each evaluation.

        If knowledge base references were provided, the retriever is created once and reused.
        Otherwise, a separate vector store is built for each embedding model, and the best-performing ones are selected.

        For evaluation only sample of the documents is used, embedded and added
        to chroma vector store.

        This method does not return anything, but in the end changes attributes
        of the instance: self.evaluation_results is a mapping holding results
        for each pattern.
        """
        logger.debug("MPS: Sampling documents")
        document_ids = []
        for element in self.benchmark_data.document_ids:
            document_ids.extend(element)

        documents = [document for document in self.documents if document.metadata["document_id"] in document_ids]
        chunked_documents = self._chunk_documents(documents)

        for i, embedding_model in enumerate(self.embedding_models):
            try:
                collection_name = f"mps_collection_{i}"
                try:
                    vector_store = self._create_vector_store(
                        embedding_model, chunked_documents, collection_name=collection_name
                    )
                except Exception as exc:
                    raise IndexingError(exc, collection_name, embedding_model.model_id) from exc

                retriever = Retriever(vector_store, **self.retrieval_params)
                self._evaluate_foundation_models(retriever=retriever, embedding_model=embedding_model)

            except IndexingError as exc:
                self._exception_handler.handle_exception(exc)
                logger.warning("Pre-evaluation of '%s' has failed.", embedding_model.model_id)
                continue

        if not self.evaluation_results:
            msg = self._exception_handler.get_final_error_msg()
            raise PreSelectorError(
                f"Foundation models pre-selection has failed. "
                f"None of the given models has been successfully evaluated. {msg}"
            )

    def _evaluate_foundation_models(self, retriever: Retriever, embedding_model: BaseEmbeddingModel):
        """
        Evaluates each embedding model with given retriever.

        Parameters
        ----------
        retriever : Retriever
            Instance to be used in retrieval phase.

        embedding_model : BaseEmbeddingModel
            Embedding model used for collection creation.
        """
        for foundation_model in self.foundation_models:
            try:
                logger.info(
                    "Starting pre-evaluation of foundation model: %s and embedding model: %s.",
                    foundation_model.model_id,
                    embedding_model.model_id,
                )

                result_scores = self._evaluate_single_pattern(foundation_model=foundation_model, retriever=retriever)

                self.evaluation_results.append(
                    {
                        "embedding_model": embedding_model,
                        "foundation_model": foundation_model,
                        "scores": result_scores.get("scores", {}),
                        "question_scores": result_scores.get("question_scores", {}),
                    }
                )

                logger.debug(
                    "Finished pre-evaluation of foundation model: %s and embedding model: %s",
                    foundation_model.model_id,
                    embedding_model.model_id,
                )
            except (GenerationError, EvaluationError) as exc:
                self._exception_handler.handle_exception(exc)
                logger.warning("Pre-evaluation of '%s' has failed.", foundation_model.model_id)
                continue

    @staticmethod
    def _create_vector_store(
        embedding_model: BaseEmbeddingModel,
        chunked_documents: list[Document],
        collection_name: str,
    ) -> BaseVectorStore:
        """
        Create instance of vector store with given chunked documents and embedding model.

        Parameters
        ----------
        embedding_model : BaseEmbeddingModel
            Embedding model used for collection creation.

        chunked_documents : list[Document]
            Chunked documents fot the embedding process.

        collection_name : str
            Name of the collection in the chroma vector database.

        Returns
        -------
        VectorStore
            Instance for communication with properly created index in the
            vector database.

        Raises
        ------
        PreSelectorError
            When 2 attempts of embedding documents are failing
        """
        logger.info("Building index for pre-evaluation using embedding model: '%s'.", embedding_model.model_id)

        vector_store = ChromaVectorStore(
            embedding_model=embedding_model,
            collection_name=collection_name,
        )

        logger.debug("MPS: Embedding documents ...")
        try:
            vector_store.add_documents(chunked_documents)
        except Exception as err:  # pylint: disable=broad-exception-caught
            logger.warning("Failed to create in-memory vector index due to: %s.", repr(err), exc_info=True)
            try:
                vector_store.add_documents(chunked_documents)
            except Exception as exc:  # pylint: disable=broad-exception-caught
                raise PreSelectorError(f"Failed to create in-memory vector index due to: {repr(exc)}.") from exc
        logger.debug("MPS: Embedding documents finished!")

        return vector_store

    def _evaluate_single_pattern(self, foundation_model: BaseFoundationModel, retriever: Retriever) -> dict[str, dict]:
        """
        Perform retrieval-augmented generation and evaluate generated response.

        Parameters
        ----------
        foundation_model : BaseFoundationModel
            Model to be used for RAG.

        retriever : Retriever
            Instance for retrieving documents from vector database.

        Returns
        -------
        dict[str, dict]
            Evaluation scores per model.
        """

        rag = SimpleRAG(foundation_model=foundation_model, retriever=retriever)

        inference_response = query_rag(
            rag=rag,
            questions=list(self.benchmark_data.questions),
        )

        result_scores = self._evaluate_response(inference_response=inference_response)

        return result_scores

    def select_models(
        self,
        n_embedding_models: int = DEFAULT_N_EMBEDDING_MODELS,
        n_foundation_models: int = DEFAULT_N_FOUNDATION_MODELS,
    ) -> dict[str, list[BaseEmbeddingModel | BaseFoundationModel]]:
        """
        Select n models pairs based on evaluation scores.

        Parameters
        ----------
        n_embedding_models : int, default=2
            Amount of embedding models to be returned.

        n_foundation_models : int, default=3
            Amount of foundation models to be returned.

        Returns
        -------
        dict[str, list[str]]
            Pre-selected embedding and foundation models.
        """

        logger.info(
            "Selecting the best %s embedding models and %s foundation models.", n_embedding_models, n_foundation_models
        )
        top_models_with_scores = self._mean_based_scoring()

        embedding_models = []
        foundation_models = []
        seen = set()
        for element in top_models_with_scores:
            fm = element.get("foundation_model")
            em = element.get("embedding_model")
            if em is not None and em not in seen:
                seen.add(em)
                embedding_models.append(em)
            if fm not in seen:
                seen.add(fm)
                foundation_models.append(fm)

        ret = {
            "foundation_models": foundation_models[:n_foundation_models],
            "embedding_models": embedding_models[:n_embedding_models],
        }

        return ret

    def _mean_based_scoring(self) -> list[dict]:
        """
        Scoring models based on mean metric value for all records used
        in the evaluation.

        Returns
        -------
        list[dict]
            Models and their corresponding mean scores in descending order.
        """
        logger.debug("MPS: Starting mean-based scoring...")

        _mean_scoring_results = []

        for result in self.evaluation_results:
            mean_score = result.get("scores", {}).get(self.metric, {}).get("mean", {})
            _mean_scoring_results.append(
                {
                    "embedding_model": result.get("embedding_model"),
                    "foundation_model": result.get("foundation_model"),
                    "score": mean_score,
                }
            )

        logger.debug("MPS: Finished mean-based scoring!")

        models_with_scores = sorted(
            _mean_scoring_results,
            key=lambda x: x.get("score"),
            reverse=True,
        )

        return models_with_scores

    def _evaluate_response(self, inference_response: list[dict[str, Any]]) -> dict[str, dict]:
        """
        Evaluate response from the model based on the chosen context,
        real questions/answers/ids from the benchmark_data.

        Parameters
        ----------
        inference_response : list[dict[str, Any]]
            List of model's responses containing question, answer and used
            reference documents for each record.

        Returns
        -------
        dict[str, dict]
            Data from evaluation that is of following structure:
            data = {
                "scores": {"answer_correctness": {"mean": 0.5, "ci_low": 0.4, "ci_high": 0.6}, ...},
                "question_scores": {
                    "answer_correctness": {"q_id_0": 0.5, "q_id_1": 0.8, ...},
                    "context_correctness": {"q_id_0": 0.5, "q_id_1": 0.8, ...},
                },
            }
        """
        logger.debug("MPS: Evaluating responses...")

        eval_data = build_evaluation_data(benchmark_data=self.benchmark_data, inference_response=inference_response)
        evaluation_result = self.evaluator.evaluate_metrics(evaluation_data=eval_data, metrics=[self.metric])

        logger.debug("MPS: Responses evaluation finished!")

        return evaluation_result

    def _chunk_documents(self, documents: list[Document]) -> list[Document]:
        """
        Chunk provided documents.

        Parameters
        ----------
        documents : list[Document]
            List of LangChain.Document instance that will be chunked.

        Returns
        -------
        list[Document]
            Chunked documents.
        """
        logger.debug("MPS: Chunking documents...")
        chunker = LangChainChunker(**self.chunking_params)

        chunked_documents = chunker.split_documents(documents)

        logger.debug("MPS: Chunking documents finished!")

        return chunked_documents



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Literal

from ai4rag.rag.vector_store.base_vector_store import BaseVectorStore


class Retriever:
    """Class responsible for retrieving data from given vector store.

    Parameters
    ----------
    vector_store : BaseVectorStore
        Vector store / vector index to retrieve data from.

    method : Literal["simple"]
        Method describing how data should be retrieved.

    number_of_chunks : int
        Number of chunks to retrieve.

    search_mode : Literal["vector", "hybrid"], default="vector"
        Search mode passed to the vector store: "vector" or "hybrid".

    ranker_strategy : str | None, default=None
        Ranking strategy for hybrid search: "rrf", "weighted", or "normalized".

    ranker_k : int | None, default=None
        Parameter k for the ranking function.

    ranker_alpha : float | None, default=None
        Alpha parameter for weighted ranking strategy.
    """

    def __init__(
        self,
        vector_store: BaseVectorStore,
        number_of_chunks: int,
        method: Literal["simple"] = "simple",
        search_mode: Literal["vector", "hybrid"] = "vector",
        ranker_strategy: str | None = None,
        ranker_k: int | None = None,
        ranker_alpha: float | None = None,
    ):
        self._vector_store = vector_store
        self.method = method
        self.number_of_chunks = number_of_chunks
        self.search_mode = search_mode
        self.ranker_strategy = ranker_strategy
        self.ranker_k = ranker_k
        self.ranker_alpha = ranker_alpha

    def retrieve(self, query: str, **kwargs) -> list[dict]:
        """Retrieve relevant documents from vector store.

        Parameters
        ----------
        query : str
            question for which documents should be retrieved.

        Returns
        -------
        list[Document]
            list of documents with their metadata corresponding to the query.
        """
        _number_of_chunks = kwargs.get("number_of_chunks", self.number_of_chunks)

        return self._vector_store.search(
            query,
            k=_number_of_chunks,
            search_mode=self.search_mode,
            ranker_strategy=self.ranker_strategy,
            ranker_k=self.ranker_k,
            ranker_alpha=self.ranker_alpha,
        )



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from abc import ABC, abstractmethod
from enum import StrEnum
from typing import TypedDict

__all__ = [
    "BaseEventHandler",
    "LogLevel",
    "PatternPayload",
    "EvaluationRecord",
]


class LogLevel(StrEnum):
    """Available log levels."""

    INFO = "info"
    WARNING = "warning"
    ERROR = "error"


# ---------------------------------------------------------------------------
# TypedDicts for on_pattern_creation – payload
# ---------------------------------------------------------------------------


class MetricCI(TypedDict):
    """Aggregate score with confidence interval for a single metric."""

    mean: float
    ci_low: float | None
    ci_high: float | None


class PatternScores(TypedDict):
    """Scores section of :class:`PatternPayload`."""

    scores: dict[str, MetricCI]
    question_scores: dict[str, dict[str, float]]


class VectorStoreSettings(TypedDict):
    """Vector store configuration used by a RAG pattern."""

    datasource_type: str
    collection_name: str


class ChunkingSettings(TypedDict):
    """Chunking parameters used during indexing."""

    method: str
    chunk_size: int
    chunk_overlap: int


class EmbeddingSettings(TypedDict):
    """Embedding model configuration used during indexing."""

    model_id: str
    distance_metric: str
    embedding_params: dict


class RetrievalSettings(TypedDict, total=False):
    """Retrieval parameters. ``window_size`` and ranker fields are optional."""

    method: str
    number_of_chunks: int
    search_mode: str
    # present only when retrieval_method == "window"
    window_size: int
    # present only when search_mode == "hybrid"
    ranker_strategy: str
    ranker_k: int
    ranker_alpha: float


class GenerationSettings(TypedDict):
    """Foundation model configuration used during generation."""

    model_id: str
    context_template_text: str
    user_message_text: str
    system_message_text: str


class PatternSettings(TypedDict):
    """Full settings block of :class:`PatternPayload`."""

    vector_store: VectorStoreSettings
    chunking: ChunkingSettings
    embedding: EmbeddingSettings
    retrieval: RetrievalSettings
    generation: GenerationSettings


class PatternPayload(TypedDict):
    """Payload passed to :meth:`BaseEventHandler.on_pattern_creation`."""

    pattern_name: str
    scores: PatternScores
    execution_time: int
    final_score: float
    schema_version: str
    producer: str
    settings: PatternSettings
    iteration: int


# ---------------------------------------------------------------------------
# TypedDicts for on_pattern_creation – evaluation_results
# ---------------------------------------------------------------------------


class AnswerContext(TypedDict):
    """Single retrieved chunk with its source document."""

    text: str
    document_id: str


class EvaluationRecord(TypedDict):
    """Per-question evaluation entry in the ``evaluation_results`` list."""

    question: str
    correct_answers: list[str]
    answer: str
    answer_contexts: list[AnswerContext]
    scores: dict[str, float]


class BaseEventHandler(ABC):
    """
    Abstract class defining interface for streaming results and messages,
    to the service layer.
    """

    @abstractmethod
    def on_status_change(self, level: LogLevel, message: str, step: str | None = None) -> None:
        """
        Method called to notify about experiment's status change.

        Parameters
        ----------
        level : LogLevel
            Logging level

        message : str
            Text of streamed message

        step : str
            Currently performed step. It should be one of composition steps.
        """

    @abstractmethod
    def on_pattern_creation(
        self, payload: PatternPayload, evaluation_results: list[EvaluationRecord], **kwargs
    ) -> None:
        """
        Method called when single RAG pattern's evaluation is completed.

        Parameters
        ----------
        payload : dict
            Information about RAG pattern's location and name, calculated scores
            and message.

            Example content:

            {
                'pattern_name': 'Pattern1',
                'scores': {
                    'scores': {
                        'answer_correctness': {'mean': 0.0, 'ci_low': None, 'ci_high': None},
                        'faithfulness': {'mean': 0.0909, 'ci_low': 0.0145, 'ci_high': 0.016},
                        'context_correctness': {'mean': 0.0, 'ci_low': None, 'ci_high': None},
                    },
                    'question_scores': {
                        'answer_correctness': {'q0': 0, 'q1': 0, 'q2': 0},
                        'faithfulness': {'q0': 0.0909, 'q1': 0.0909, 'q2': 0.0909},
                        'context_correctness': {'q0': 0, 'q1': 0, 'q2': 0},
                    },
                },
                'execution_time': 0,
                'final_score': 0.0909,
                'schema_version': '1.0',
                'producer': 'ai4rag',
                'settings': {
                    'vector_store': {'datasource_type': 'chroma', 'collection_name': 'ai4rag_20260317092550'},
                    'chunking': {'method': 'recursive', 'chunk_size': 1024, 'chunk_overlap': 256},
                    'embedding': {
                        'model_id': 'mock-em-1',
                        'distance_metric': 'cosine',
                        'embedding_params': {'embedding_dimension': 64},
                    },
                    'retrieval': {'method': 'window', 'number_of_chunks': 3, 'search_mode': 'vector', 'window_size': 3},
                    'generation': {
                        'model_id': 'mock-fm-2',
                        'context_template_text': '{document}',
                        'user_message_text': '\n\nContext:\n{reference_documents}:\n\nQuestion: {question}.\n
                        'system_message_text': 'System instruction...'
                    }
                },
                'iteration': 0,
            }


        evaluation_results : dict
            Results from single pattern evaluation.

            Example content:

            [
                {
                    "question": "<question_1>"
                    "answer": "<model's answer>",
                    "answer_contexts": [
                        {"text": "<content1_text>", "document_id": "document_1.pdf"},
                        {"text": "<content2_text>", "document_id": "document_2.pdf"},
                    ]
                    'correct_answers': ['correct_answer_for_question_1'],
                    "scores": {
                        "answer_correctness": 0.79,
                        "faithfulness": 0.55,
                        "context_correctness": 0.65,
                    }
                },
                {
                    "question": "<question_2>",
                    "answer": "<model's answer>",
                    "answer_contexts": [
                        {"text": "<content3_text>", "document_id": "document_3.pdf"},
                        {"text": "<content4_text>", "document_id": "document_4.pdf"},
                    ]
                    "correct_answers": ["correct_answer_1_for_question_2", "correct_answer_2_for_question_3"],
                    "scores": {
                        "answer_correctness": 0.79,
                        "faithfulness": 0.55,
                        "context_correctness": 0.65,
                    },
                },
            ]
        """



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from abc import ABC, abstractmethod
from typing import Any, Generic, Protocol, Self, TypeVar, overload

T = TypeVar("T")
CT = TypeVar("CT", bound="Comparable")


class ConstraintsValidationError(Exception):
    """Error raised when validation has failed."""


class Comparable(Protocol):
    """Protocol which ensures that the type is comparable."""

    def __eq__(self, other: Self, /) -> bool: ...
    def __lt__(self, other: Self, /) -> bool: ...
    def __le__(self, other: Self, /) -> bool: ...
    def __gt__(self, other: Self, /) -> bool: ...
    def __ge__(self, other: Self, /) -> bool: ...


class Validator(Generic[T], ABC):
    """Base validator class."""

    def __init__(self):
        self.private_name: str

    def __set_name__(self, _, name):
        self.private_name = f"_{name}"

    def __set__(self, obj, value: T):
        validated_value = self.validate(obj, value)
        setattr(obj, self.private_name, validated_value)

    @overload
    def __get__(self, instance: None, owner: Any) -> Self: ...

    @overload
    def __get__(self, instance: Any, owner: Any) -> T: ...

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return getattr(instance, self.private_name)

    @abstractmethod
    def validate(self, obj, value):
        """Base validate method."""


class OneOf(Validator[T]):
    """Validates if given value is within provided set of values."""

    def __init__(self, *options: T):
        super().__init__()
        self.options = set(options)

    def validate(self, obj, value: T):
        if value not in self.options:
            raise ConstraintsValidationError(
                f"Expected {value!r} to be one of {self.options!r} for {type(obj).__name__} "
                f"on attribute {self.private_name[1:]}."
            )
        return value



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from typing import Any, Sequence

from ai4rag.utils.constants import ConstantMeta


# pylint: disable=too-many-instance-attributes
@dataclass
class EvaluationData:
    """
    Representation of data sent for evaluation.

    Parameters
    ----------
    question : str | None, default=None
        Original question from the benchmark dataset

    answer : str | None, default=None
        Answer returned by the model.

    contexts : list[str] | None = None
        Contexts used by the model to generate response.

    context_ids: list[str] | None, default=None
        IDs of contexts used by the model to generate response.

    ground_truths : list[str] | None = None
        Correct answers from the benchmark dataset.

    ground_truths_context_ids : list[str] | None = None
        IDs of the correct documents used for answers in the benchmark dataset.

    question_id : str | None = None
        ID of the question.

    additional_data: dict[str, Any] | None = None
        Any additional data associated with the evaluation results.

    Methods
    -------
    to_dict() -> dict[str, Any]
        Used for casting instance to the dictionary
    """

    question: str | None = None
    answer: str | None = None
    contexts: list[str] | None = None
    context_ids: list[str] | None = None
    ground_truths: list[str] | None = None
    ground_truths_context_ids: list[str] | None = None
    question_id: str | None = None
    additional_data: list[Any] | None = None

    def to_dict(self) -> dict[str, Any]:
        """Cast given instance of dataclass to the dict."""
        return asdict(self)


class MetricType(metaclass=ConstantMeta):
    """
    Holder for metric names.

    Attributes
    ----------
    ANSWER_CORRECTNESS : str, default="answer_correctness"

    FAITHFULNESS : str, default="faithfulness"

    CONTEXT_CORRECTNESS : str, default="context_correctness"
    """

    ANSWER_CORRECTNESS = "answer_correctness"
    FAITHFULNESS = "faithfulness"
    CONTEXT_CORRECTNESS = "context_correctness"


class BaseEvaluator(ABC):
    """
    This class defines the functionality to evaluate a RAG application
    and compare different RAG applications.
    """

    @abstractmethod
    def evaluate_metrics(self, evaluation_data: list[EvaluationData], metrics: Sequence[str]) -> dict:
        """
        Evaluate the model's responses against list of different metrics.

        Parameters
        ----------
        evaluation_data : list[EvaluationData]
            List of EvaluationData instances containing all the data needed
            to perform evaluation.

        metrics : Sequence[str]
            Metrics given as strings. They should be referred to MetricType.

        Returns
        -------
        dict
            Evaluation result data.
        """



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Literal

from ai4rag.utils.constants import ConstantMeta

__all__ = ["EmbeddingModels", "FoundationModels"]

DEFAULT_MAX_SEQUENCE_LENGTH = 8192


class Models(metaclass=ConstantMeta):
    """Parent class from models classes that is an instance of ConstantMeta."""

    @classmethod
    def get_models(cls) -> list[str]:
        """
        Get list of supported models from class attributes.

        Returns
        -------
        list[str]
            Supported models
        """
        # pylint: disable=no-member
        models = list(el for el in cls._constant_attributes)

        return models


class EmbeddingModels(Models):
    """
    Available embedding models on watsonx.ai. Similar structure is supported
    by FoundationModelsManager.EmbeddingModels(), but this is hardcoded here, so
    we do not need to create api client for reading models.

    List of supported embedding models can be found here:
    https://dataplatform.cloud.ibm.com/docs/content/wsj/analyze-data/fm-models-embed.html?context=wx
    """

    IBM_SLATE_125M_ENG = "ibm/slate-125m-english-rtrvr"
    IBM_SLATE_125M_ENG_V2 = "ibm/slate-125m-english-rtrvr-v2"
    IBM_GRANITE_278M_MULTILINGUAL = "ibm/granite-embedding-278m-multilingual"
    MULTILINGUAL_E5_LARGE = "intfloat/multilingual-e5-large"

    @classmethod
    def get_max_tokens(cls, model_id: str) -> int:
        """
        This function returns the maximum number of tokens that can be passed
        to the embed function for a given model ID. If no match is found in the map,
        a default value of 512 is returned.

        Parameters
        ----------
        model_id : str
            The model ID for which to retrieve the maximum number of input tokens.

        Returns
        -------
        int
            The maximum number of input tokens for the given model ID. Default is 512.
        """
        max_input_tokens_map = {
            cls.IBM_SLATE_125M_ENG: 512,
            cls.IBM_SLATE_125M_ENG_V2: 512,
            cls.IBM_GRANITE_278M_MULTILINGUAL: 512,
            cls.MULTILINGUAL_E5_LARGE: 512,
        }

        return max_input_tokens_map.get(model_id, 512)

    @classmethod
    def get_distance_metric(cls, model_id: str) -> Literal["cosine", "euclidean"]:
        """
        This function returns the appropriate distance metric to be used for
         constructing a vector store and retrieving passages for given model id.

        Parameters
        ----------
        model_id : str
            The model ID for which to retrieve the maximum number of input tokens.

        Returns
        -------
        str
            Distance metric to be used for constructing a vector store and retrieving.
        """
        distance_metric_map = {
            cls.IBM_SLATE_125M_ENG: "cosine",
            cls.MULTILINGUAL_E5_LARGE: "cosine",
            cls.IBM_SLATE_125M_ENG_V2: "cosine",
            cls.IBM_GRANITE_278M_MULTILINGUAL: "cosine",
        }
        return distance_metric_map.get(model_id, "cosine")


class FoundationModels(Models):
    """Available foundation models."""

    MISTRAL_SMALL_3_1_24B_INSTRUCT = "mistralai/mistral-small-3-1-24b-instruct-2503"
    MISTRAL_MISTRAL_LARGE = "mistralai/mistral-large"
    MISTRAL_MEDIUM_2505 = "mistralai/mistral-medium-2505"
    GRANITE_3_8B_INSTRUCT = "ibm/granite-3-8b-instruct"
    GRANITE_3_3_8B_INSTRUCT = "ibm/granite-3-3-8b-instruct"
    GRANITE_4H_SMALL = "ibm/granite-4-h-small"
    META_LLAMA_3_1_70B_INSTRUCT = "meta-llama/llama-3-1-70b-instruct"
    META_LLAMA_3_1_8B_INSTRUCT = "meta-llama/llama-3-1-8b-instruct"
    META_LLAMA_3_3_70B_INSTRUCT = "meta-llama/llama-3-3-70b-instruct"
    META_LLAMA_4_MAVERICK_17B_128E_INSTRUCT_FP8 = "meta-llama/llama-4-maverick-17b-128e-instruct-fp8"
    OPENAI_GPT_OSS_120B = "openai/gpt-oss-120b"

    @classmethod
    def get_default_max_sequence_length(cls, model_id: str, default: int = DEFAULT_MAX_SEQUENCE_LENGTH) -> int:
        """
        Get default max sequence length specific to model. This data should be
        updated as this is some mapping made based on the models hosted on watsonx.
        It is introduced as a response to unstable endpoints containing 'model_limits'.

        Parameters
        ----------
        model_id : str
            Model ID

        default : int, default=8192
            Default value to be returned when no value is set for the model.

        Returns
        -------
        int
            Max sequence length for a specific model. Default is 8192.
        """

        defaults = {
            cls.MISTRAL_MISTRAL_LARGE: 128_000,
            cls.MISTRAL_SMALL_3_1_24B_INSTRUCT: 131_072,
            cls.MISTRAL_MEDIUM_2505: 131_072,
            cls.META_LLAMA_3_1_70B_INSTRUCT: 128_000,
            cls.META_LLAMA_3_1_8B_INSTRUCT: 128_000,
            cls.META_LLAMA_3_3_70B_INSTRUCT: 131_072,
            cls.META_LLAMA_4_MAVERICK_17B_128E_INSTRUCT_FP8: 131_072,
            cls.GRANITE_3_8B_INSTRUCT: 131_072,
            cls.GRANITE_3_3_8B_INSTRUCT: 131_072,
            cls.OPENAI_GPT_OSS_120B: 131_072,
            cls.GRANITE_4H_SMALL: 131_072,
        }

        return defaults.get(model_id, default)



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import TypedDict

from llama_stack_client import LlamaStackClient

from ai4rag import logger
from ai4rag.rag.embedding.llama_stack import LSEmbeddingModel, LSEmbeddingParams
from ai4rag.rag.foundation_models.llama_stack import LSFoundationModel
from ai4rag.search_space.src.exceptions import SearchSpaceValueError


class _DefaultModelsResponseType(TypedDict):
    foundation_models: list[LSFoundationModel]
    embedding_models: list[LSEmbeddingModel]


def _validate_foundation_model(model: LSFoundationModel) -> bool:
    """
    Validate that a foundation model responds correctly with minimal tokens.

    Parameters
    ----------
    model : LSFoundationModel
        Foundation model to validate.

    Returns
    -------
    bool
        True if model responds successfully, False otherwise.
    """
    try:
        # Test with minimal message and 1 token response
        test_messages = [{"role": "user", "content": "Hi"}]
        model.chat(messages=test_messages)
        return True
    except Exception:  # pylint: disable=broad-exception-caught
        logger.warning(
            "Foundation model '%s' does not respond and will be excluded from search space.",
            model.model_id,
            exc_info=True,
        )
        return False


def _validate_embedding_model(model: LSEmbeddingModel) -> bool:
    """
    Validate that an embedding model responds correctly with minimal input.

    Parameters
    ----------
    model : LSEmbeddingModel
        Embedding model to validate.

    Returns
    -------
    bool
        True if model responds successfully, False otherwise.
    """
    try:
        # Test with minimal text
        model.embed_query("test")
        return True
    except Exception:  # pylint: disable=broad-exception-caught
        logger.warning(
            "Embedding model '%s' does not respond and will be excluded from search space.",
            model.model_id,
            exc_info=True,
        )
        return False


def _get_default_llama_stack_models(client: LlamaStackClient) -> _DefaultModelsResponseType:
    """Get list of default foundation models based on the available ones in llama stack."""

    logger.info("Selecting default foundation models...")
    available_models = client.models.list()
    llms = [model for model in available_models if model.custom_metadata.get("model_type") == "llm"]
    embeddings = [model for model in available_models if model.custom_metadata.get("model_type") == "embedding"]

    # Create model instances
    foundation_models_unvalidated = [LSFoundationModel(model_id=m.id, client=client) for m in llms]
    embedding_models_unvalidated = [
        LSEmbeddingModel(
            model_id=m.id,
            client=client,
            params=LSEmbeddingParams(
                embedding_dimension=getattr(m, "custom_metadata", {}).get("embedding_dimension"),
                context_length=getattr(m, "custom_metadata", {}).get("context_length"),
            ),
        )
        for m in embeddings
    ]

    # Validate each model
    logger.info("Validating foundation models...")
    foundation_models = [m for m in foundation_models_unvalidated if _validate_foundation_model(m)]

    logger.info("Validating embedding models...")
    embedding_models = [m for m in embedding_models_unvalidated if _validate_embedding_model(m)]

    if not foundation_models:
        raise SearchSpaceValueError("There are no available models of type 'llm'.")
    if not embedding_models:
        raise SearchSpaceValueError("There are no available models of type 'embedding'.")

    logger.info("Selected default foundation models: %s.", foundation_models)
    logger.info("Selected default embedding models: %s.", embedding_models)

    return {"foundation_models": foundation_models, "embedding_models": embedding_models}


def _are_provided_models_available(
    provided_models: list, available_models: list[LSFoundationModel | LSEmbeddingModel]
) -> bool:
    """Check whether models provided by the user are available for the experiment."""

    available_ids = [m.model_id for m in available_models]

    for model in provided_models:
        m_id = model.model_id
        if m_id not in available_ids:
            raise SearchSpaceValueError(f"Provided model with model_id: '{m_id}' is not available for the experiment.")
    return True



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from collections import defaultdict

from langchain_core.documents import Document


def merge_window_into_a_document(window: list[Document]) -> Document:
    """
    Merges a list of chunks into a single document.
    If consecutive chunks have intersecting merged_text, the merged_text is merged to avoid duplications.

    Parameters
    ----------
    window : list[Document]
        Ordered list of documents for merging.

    Returns
    -------
    Document
        object that contains the merged merged_text of the window documents.
    """

    def merge_metadata(multiple_metadata: list[dict]) -> dict:
        """
        Merges a list of dictionaries (metadata) into one metadata.
        The keys remain the same but the values are changed into lists of values from all metadata.

        Parameters
        ----------
        multiple_metadata : list[dict]
            List of metadata dictionaries to be merged.

        Returns
        -------
        Single
            Merged dictionary (metadata) with unique values in lists for each key.
        """
        if len(multiple_metadata) == 1:
            return multiple_metadata[0]

        merged_metadata = defaultdict(set)
        for metadata in multiple_metadata:
            for key, value in metadata.items():
                if isinstance(value, list):
                    merged_metadata[key].update(value)
                else:
                    merged_metadata[key].add(value)

        result = {}
        for key, value_set in merged_metadata.items():
            value_list = sorted(value_set)
            if len(value_list) == 1:
                result[key] = value_list[0]
            else:
                result[key] = value_list
        return result

    def get_str2_without_intersecting_text(str1: str, str2: str) -> tuple[str, bool]:
        """
        Finds the intersecting merged_text between the suffix of str1 and the prefix of str2.

        Parameters
        ----------
        str1 : str
            The first string.

        str2 : str
            The second string.

        Returns
        -------
        tuple[str, bool]
            1. str2 without its intersection to str1
            2. whether there was an intersection or not
        """
        # Start checking from the longest possible overlap to the shortest
        for i in range(min(len(str1), len(str2)), 0, -1):
            if str1[-i:] == str2[:i]:
                return str2[i:], True
        return str2, False

    def merge_texts(texts: list[str]) -> str:
        """
        Merges a list of texts into a single text string.
        If consecutive texts have intersecting parts, the text is merged to avoid duplications.

        Parameters
        ----------
        texts : list[str]
            Ordered list of text strings to be merged.

        Returns
        -------
        str
            Single string that contains the merged text.
        """
        merged_text = ""
        for text in texts:
            text_to_add, has_intersection = get_str2_without_intersecting_text(merged_text, text)
            if merged_text and not has_intersection:
                merged_text += " "  # Add a space between non-overlapping texts (chunks)
            merged_text += text_to_add
        return merged_text

    texts = [doc.page_content for doc in window]
    merged_text = merge_texts(texts)

    metadata = [doc.metadata for doc in window]
    merged_metadata = merge_metadata(metadata)

    return Document(page_content=merged_text, metadata=merged_metadata)



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from typing import Any, Literal, TypeAlias, TypedDict, TypeVar

from ai4rag import logger
from ai4rag.core.experiment.benchmark_data import BenchmarkData
from ai4rag.core.experiment.exception_handler import GenerationError
from ai4rag.evaluator.base_evaluator import EvaluationData
from ai4rag.rag.embedding.base_model import BaseEmbeddingModel
from ai4rag.rag.foundation_models.base_model import BaseFoundationModel
from ai4rag.rag.template.base_template import BaseRAGTemplate
from ai4rag.utils.constants import AI4RAGParamNames

T = TypeVar("T")

__all__ = [
    "VectorStoreType",
    "RAGExperimentError",
    "RAGParamsType",
    "query_rag",
    "build_evaluation_data",
    "get_retrieval_params",
    "get_chunking_params",
    "RAGRetrievalParamsType",
]

_semantic_chunker_cache = {}

VectorStoreType: TypeAlias = Literal["milvus", "chroma"]


class RAGExperimentError(Exception):
    """Exception representing error in the experiment."""


class RAGParamsType(TypedDict):
    """Parameters required for single AutoRAG Pattern evaluation."""

    embedding_model: BaseEmbeddingModel
    foundation_model: BaseFoundationModel
    chunk_size: int
    chunk_overlap: int | float
    chunking_method: Literal["recursive"]
    window_size: int
    number_of_chunks: int
    retrieval_method: Literal["simple", "window"]
    search_mode: Literal["vector", "hybrid"]
    ranker_strategy: str
    ranker_k: int
    ranker_alpha: int | float


class RAGChunkingParamsType(TypedDict):
    """Required chunking parameters."""

    chunk_size: int
    chunk_overlap: int | float
    chunking_method: Literal["recursive"]


class RAGRetrievalParamsType(TypedDict):
    """Required retrieval parameters."""

    retrieval_window_size: int
    number_of_retrieved_chunks: int
    retrieval_method: Literal["simple", "window"]
    search_mode: Literal["vector", "hybrid"]
    ranker_strategy: str
    ranker_k: int
    ranker_alpha: int | float


def query_rag(rag: BaseRAGTemplate, questions: list[str], max_threads: int = 10) -> list[dict[str, Any]]:
    """
    Function to perform parallel queries on RAG inference service.

    Parameters
    ----------
    rag : BaseRAGTemplate
        Instance of the BaseRAGTemplate to be used for retrieval-augmented generation.

    questions : list[str]
        Questions used for AI Service (RAG).

    max_threads : int, default=10
        Limit of the concurrent workers querying the AI service. Value of 10
        is chosen based on the client's inference endpoint settings.

    Returns
    -------
    list[dict[str, Any]]
        List of dicts as in the _generate_response.
    """
    logger.debug(
        "Starting concurrent RAG execution. Limit of concurrent executions: %s for %s calls. Model: %s",
        max_threads,
        len(questions),
        rag.foundation_model.model_id,
    )

    try:
        _generate_function = partial(_generate_response, rag=rag)

        with ThreadPoolExecutor(max_workers=max_threads) as executor:
            responses = list(executor.map(_generate_function, questions))

    except Exception as exc:
        raise GenerationError(exc, model_id=rag.foundation_model.model_id) from exc

    logger.debug("Finished concurrent RAG execution!")

    return responses


def _generate_response(question: str, rag: BaseRAGTemplate) -> dict[str, Any]:
    """
    Make a single call to the RAG instance.
    Notice that question parameter should remain first to be easily
    utilized by concurrent executor.

    Parameters
    ----------
    question : str
        Question for the RAG.

    rag : BaseRAGTemplate
        Instance capable of performing Retrieval-Augmented Generation.

    Returns
    -------
    dict[str, Any]
        Example result:
        {
            "question": "What is the meaning of life?",
            "answer": "Being good to other people."
            "reference_documents": [
                {"page_content": "Document content 1", "metadata": {"document_id": "doc_id_1", ...}},
                {"page_content": "Document content 2", "metadata": {"document_id": "doc_id_2", ...}},
                ...,
            ]
        }
    """
    return rag.generate(question)


def build_evaluation_data(
    benchmark_data: BenchmarkData, inference_response: list[dict[str, Any]]
) -> list[EvaluationData]:
    """
    Helper function responsible for building payload for response evaluation.

    Parameters
    ----------
    benchmark_data : BenchmarkData
        Instance holding information about questions, answers and ids.

    inference_response : list[dict[str, Any]]
        List of model's responses containing question, answer and used
        reference documents for each record.

    Returns
    -------
    list[EvaluationData]
        Sequence containing data that will be used for evaluation.
    """
    evaluation_data = []

    for idx in range(len(benchmark_data)):
        contexts = []
        context_ids = []
        for el in inference_response[idx]["reference_documents"]:
            contexts.append(getattr(el, "page_content", None))
            context_ids.append(getattr(el, "metadata", {}).get("document_id"))

        evaluation_data.append(
            EvaluationData(
                question=benchmark_data.questions[idx],
                answer=inference_response[idx]["answer"],
                contexts=contexts,
                context_ids=context_ids,
                ground_truths=benchmark_data.correct_answers[idx],
                question_id=benchmark_data.questions_ids[idx],
                ground_truths_context_ids=benchmark_data.document_ids[idx] if benchmark_data.document_ids else None,
            )
        )

    return evaluation_data


def _get_chunk_overlap(chunk_size: int, chunk_overlap: int | float) -> int:
    """
    Get chunking overlap as number of tokens/characters used as cross-chunk overlap

    Parameters
    ----------
    chunk_size : int
        Size of the created chunks.

    chunk_overlap : int | float
        If "int", the chunk_overlap is considered as number of token/characters.
        If "float", it's expected to be in the range [0, 1] and it'll be treated as a
        percentage of the chunk_size.

    Returns
    -------
    int
        number of characters/tokens used as overlap between chunks
    """
    if isinstance(chunk_overlap, float):
        if chunk_overlap is None or (chunk_overlap < 0 or chunk_overlap > 1):
            raise ValueError(
                "chunk_overlap is expected to be an integer >= 0 or a floating-point number between 0 and 1."
            )
        chunk_overlap = int(chunk_size * chunk_overlap)
    return chunk_overlap


def get_chunking_params(rag_params: RAGParamsType) -> RAGChunkingParamsType:
    """
    Extracts chunking parameters from the provided rag parameters.
    All three configurations are mandatory as part of single `chunking` setting:
        `method`, `chunk_size`, `chunk_overlap`

    Parameters
    ----------
    rag_params : RAGParamsType
        Dictionary with chunking setting for single evaluation run.

    Returns
    -------
    dict
        Dictionary with chunking parameters: chunking_method, chunk_size, chunk_overlap.

    Raises
    ------
    RAGExperimentError
        Raised when chunking parameters are missing.
    """
    chunking_params = {
        k: rag_params.get(k)
        for k in [AI4RAGParamNames.CHUNKING_METHOD, AI4RAGParamNames.CHUNK_SIZE, AI4RAGParamNames.CHUNK_OVERLAP]
    }
    chunking_params[AI4RAGParamNames.CHUNK_OVERLAP] = _get_chunk_overlap(
        chunking_params[AI4RAGParamNames.CHUNK_SIZE], chunking_params[AI4RAGParamNames.CHUNK_OVERLAP]
    )
    if any(v is None for v in chunking_params.values()):
        raise RAGExperimentError(f"Missing or invalid values in chunking configuration: {chunking_params}.")

    return chunking_params


def get_retrieval_params(rag_params: RAGParamsType) -> RAGRetrievalParamsType:
    """
    Extracts retrieval parameters from the provided rag parameters.
    All three setting's configurations are mandatory under `retrieval` key:
        `method`, `window_size`, `number_of_chunks`

    Parameters
    ----------
    rag_params : RAGParamsType
        Dictionary with retrieval setting for single evaluation run.

    Returns
    -------
    RAGRetrievalParamsType
        retrieval_method, retrieval_window_size, number_of_retrieved_chunks, search_mode.

    Raises
    ------
    RAGExperimentError
        Raised when retrieval parameters are missing.
    """
    retrieval_params = {
        AI4RAGParamNames.WINDOW_SIZE: rag_params.get(AI4RAGParamNames.WINDOW_SIZE),
        AI4RAGParamNames.NUMBER_OF_CHUNKS: rag_params.get(AI4RAGParamNames.NUMBER_OF_CHUNKS),
        AI4RAGParamNames.SEARCH_MODE: rag_params.get(AI4RAGParamNames.SEARCH_MODE),
        AI4RAGParamNames.RETRIEVAL_METHOD: rag_params.get(AI4RAGParamNames.RETRIEVAL_METHOD),
        AI4RAGParamNames.RANKER_STRATEGY: rag_params.get(AI4RAGParamNames.RANKER_STRATEGY),
        AI4RAGParamNames.RANKER_K: rag_params.get(AI4RAGParamNames.RANKER_K),
        AI4RAGParamNames.RANKER_ALPHA: rag_params.get(AI4RAGParamNames.RANKER_ALPHA),
    }
    required_keys = (
        AI4RAGParamNames.WINDOW_SIZE,
        AI4RAGParamNames.NUMBER_OF_CHUNKS,
        AI4RAGParamNames.SEARCH_MODE,
        AI4RAGParamNames.RETRIEVAL_METHOD,
    )
    if any(retrieval_params.get(k) is None for k in required_keys):
        raise RAGExperimentError(f"Missing or invalid values in retrieval configuration: {retrieval_params}.")

    return retrieval_params



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Sequence

import pandas as pd
from unitxt.eval_utils import evaluate

from ai4rag.core.experiment.exception_handler import EvaluationError
from ai4rag.evaluator.base_evaluator import (
    BaseEvaluator,
    EvaluationData,
    MetricType,
)


class UnitxtEvaluator(BaseEvaluator):
    """Unitxt wrapper making evaluation of the RAG's usage."""

    METRIC_TYPE_MAP = {
        MetricType.ANSWER_CORRECTNESS: "metrics.rag.external_rag.answer_correctness",
        MetricType.FAITHFULNESS: "metrics.rag.external_rag.faithfulness",
        MetricType.CONTEXT_CORRECTNESS: "metrics.rag.external_rag.context_correctness",
    }

    def evaluate_metrics(
        self,
        evaluation_data: list[EvaluationData],
        metrics: Sequence[str],
    ) -> dict:
        """
        Perform evaluation on the given instances with chosen metric types.

        Parameters
        ----------
        evaluation_data : list[EvaluationData]
            Instances that hold data needed for the unitxt algorithms to perform evaluation.

        metrics : Sequence[str]
            Values describing which specific evaluation metrics should be used
            withing evaluation process.

        Returns
        -------
        dict
            Dictionary of scores given for each EvaluationData.
        """
        evaluation_primitives = [prim.to_dict() for prim in evaluation_data]
        df = pd.DataFrame(evaluation_primitives)
        unitxt_metrics = self.get_metric_types(metric_types=metrics)
        try:
            scores_df, ci_table = evaluate(df, metric_names=unitxt_metrics, compute_conf_intervals=True)

            returned_ci = self._handle_ci_calculations(ci_table=ci_table)
            question_scores = self._handle_questions_scores(scores_df=scores_df)

            return {"scores": returned_ci, "question_scores": question_scores}

        except Exception as exc:
            raise EvaluationError(exc) from exc

    def _handle_questions_scores(self, scores_df: pd.DataFrame) -> dict:
        """
        Handle transformations of questions scores data frame.

        Parameters
        ----------
        scores_df : pd.DataFrame
            Data returned by the unitxt evaluate function containing scores without ids.

        Returns
        -------
        dict
            Scores calculated for each question in the evaluation data.
        """
        reversed_metrics_mapping = {v: k for k, v in self.METRIC_TYPE_MAP.items()}
        scores_df = scores_df.mask(scores_df == "")
        raw_ret_dict = scores_df.round(4).to_dict()
        without_id = {
            reversed_metrics_mapping[key]: val for key, val in raw_ret_dict.items() if key in reversed_metrics_mapping
        }
        question_scores = {}
        for key, val in without_id.items():
            question_scores[key] = {raw_ret_dict["question_id"][k]: v for k, v in val.items()}

        return question_scores

    def _handle_ci_calculations(self, ci_table: pd.DataFrame) -> dict:
        """
        Handle transformations of confidence interval scores data frame.

        Parameters
        ----------
        ci_table : pd.DataFrame
            Data with calculated confidence intervals via unitxt evaluate.

        Returns
        -------
        dict
            Transformed confidence interval data that will be further processed.
        """
        reversed_metrics_mapping = {v: k for k, v in self.METRIC_TYPE_MAP.items()}
        ci_dict = ci_table.to_dict()

        def round_or_none(x: float | None) -> float | None:
            return None if x is None or pd.isna(x) else round(x, 4)

        returned_ci = {}
        for key, val in ci_dict.items():
            returned_ci[reversed_metrics_mapping[key]] = {
                "mean": round_or_none(val["score"]),
                "ci_low": round_or_none(val.get("score_ci_low")),
                "ci_high": round_or_none(val.get("score_ci_high")),
            }

        return returned_ci

    @classmethod
    def get_metric_types(cls, metric_types: Sequence[str]) -> list[str]:
        """
        Perform mapping of general metric names to the specific metric names
        in the unitxt library.

        Parameters
        ----------
        metric_types : Sequence[str]
            Metrics defined in the MetricType class.

        Returns
        -------
        list[str]
            Specific versions of the metrics that can be used within
            unitxt evaluation process.
        """
        mapping = [cls.METRIC_TYPE_MAP.get(metric, None) for metric in metric_types]
        return [metric for metric in mapping if metric is not None]

    @classmethod
    def decode_unitxt_metric(cls, unitxt_metrics: list[str]) -> list[str]:
        """
        Decode metrics from the unitxt names to general names.

        Parameters
        ----------
        unitxt_metrics : list[str]
            Encoded unitxt metrics.

        Returns
        -------
        list[str]
            Corresponding decoded messages
        """

        reversed_mapping = {v: k for k, v in cls.METRIC_TYPE_MAP.items()}
        decoded = [reversed_mapping[metric] for metric in unitxt_metrics]

        return decoded



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from ai4rag.search_space.src.parameter import Parameter
from ai4rag.utils.constants import AI4RAGParamNames

__all__ = [
    "get_default_ai4rag_search_space_parameters",
]

# Note: "" and 0 are sentinels for unused params; ranker_alpha uses 1 as sentinel (0 means 100% sparse)
_default_chunking_methods = ("recursive",)
_default_chunk_sizes = (512, 1024, 2048, 4096)
_default_chunk_overlaps = (128, 256, 512)
_default_retrieval_methods = ("simple",)
_default_window_sizes = (0,)
_default_chroma_retrieval_methods = ("simple", "window")
_default_chroma_window_sizes = (0, 1, 3, 5)
_default_numbers_of_chunks = (3, 5, 10)
_default_search_modes = ("vector", "hybrid")  # currently off as llama stack has issue with hybrid search mode
_default_ranker_strategies = ("", "rrf", "weighted")  # to extend with normalized
_default_ranker_k = (0, 60)  # currently off as llama stack has issue with hybrid search mode
_default_ranker_alpha = (1, 0.5)  # currently off as llama stack has issue with hybrid search mode


def get_default_ai4rag_search_space_parameters(vector_store_type: str = "ls_milvus") -> list[Parameter]:
    """
    Function to return default search space containing experiment parameters.

    Parameters
    ----------
    vector_store_type : str, default="ls_milvus"
        Type of vector store. When "chroma", hybrid search parameters are excluded
        since ChromaDB does not support hybrid search.

    Returns
    -------
    list[Parameter]
        Parameters that will be used for creating AI4RAGSearchSpace
    """

    if vector_store_type == "chroma":
        retrieval_methods = _default_chroma_retrieval_methods
        window_sizes = _default_chroma_window_sizes
    else:
        retrieval_methods = _default_retrieval_methods
        window_sizes = _default_window_sizes

    default_search_space_parameters = [
        Parameter(name=AI4RAGParamNames.CHUNKING_METHOD, values=_default_chunking_methods),
        Parameter(name=AI4RAGParamNames.CHUNK_SIZE, values=_default_chunk_sizes),
        Parameter(name=AI4RAGParamNames.CHUNK_OVERLAP, values=_default_chunk_overlaps),
        Parameter(name=AI4RAGParamNames.RETRIEVAL_METHOD, values=retrieval_methods),
        Parameter(name=AI4RAGParamNames.WINDOW_SIZE, values=window_sizes),
        Parameter(name=AI4RAGParamNames.NUMBER_OF_CHUNKS, values=_default_numbers_of_chunks),
    ]

    if vector_store_type == "chroma":
        default_search_space_parameters.append(
            Parameter(name=AI4RAGParamNames.SEARCH_MODE, values=("vector",)),
        )
    else:
        default_search_space_parameters.extend(
            [
                Parameter(name=AI4RAGParamNames.SEARCH_MODE, values=("vector",)),
            ]
        )

    return default_search_space_parameters



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
import logging
import os

__version__ = "0.5.0"

logger = logging.getLogger("ai4rag")
logger.setLevel(os.getenv("LOG_LEVEL", "INFO"))

formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

handler = logging.StreamHandler()
handler.setFormatter(formatter)
logger.addHandler(handler)



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from collections import Counter

from ai4rag import logger
from ai4rag.utils.event_handler import BaseEventHandler, LogLevel


class AI4RAGError(Exception):
    """Base class for all external exceptions which can occur in AI4RAG."""

    def __init__(self, exception: Exception, message: str | None = None) -> None:
        self.exception = exception
        self.message = message

    def __repr__(self) -> str:
        if self.message:
            return f"AI4RAGError: {self.message} due to {repr(self.exception)}"
        return f"{self.__class__.__name__}: {repr(self.exception)})"

    def __str__(self) -> str:
        return repr(self)


class IndexingError(AI4RAGError):
    """Exception representing error during indexing chunks to vector store."""

    def __init__(self, exception, collection_name, embedding_model_id):
        super().__init__(exception)
        self.collection_name = collection_name
        self.embedding_model_id = embedding_model_id

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}: Unable to embed and upload documents to vector store for collection name "
            f"'{self.collection_name}' and embedding model '{self.embedding_model_id}' due to: {repr(self.exception)}"
        )


class GenerationError(AI4RAGError):
    """Exception representing error during retrieval or inference."""

    def __init__(self, exception: Exception, model_id: str):
        super().__init__(exception)
        self.model_id = model_id
        self._message_core = f"Unable to retrieve chunks and generate answers for foundation model: '{model_id}'"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}: {self._message_core} due to: {repr(self.exception)}"


class EvaluationError(AI4RAGError):
    """Exception representing error during evaluating generated pattern"""

    def __init__(self, exception):
        super().__init__(exception)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}: Unable to evaluate generated pattern due to: {repr(self.exception)}"


class AssetSaveError(AI4RAGError):
    """Exception representing error during writing finished pattern to COS instance."""

    def __init__(self, exception):
        super().__init__(exception)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}: Unable to save assets for evaluated pattern due to: {repr(self.exception)}"


class ExperimentExceptionHandler:
    """
    Class responsible for handling exceptions raised during the experiment.
    Exceptions are stored in dict containing information about iteration and raised error.
    """

    def __init__(self, event_handler: BaseEventHandler | None = None) -> None:
        self.errors = []
        self.event_handler = event_handler

    def handle_exception(self, exception: AI4RAGError):
        """
        Handles single exception raised during the experiment.

        Parameters
        ----------
        exception : AI4RAGError
            Handled exception.
        """
        self.errors.append(exception)
        logger.warning(exception, exc_info=True)
        msg = repr(exception)
        if self.event_handler:
            self.event_handler.on_status_change(level=LogLevel.WARNING, message=msg)

        return msg

    def get_final_error_msg(self) -> str:
        """
        Counts exceptions raised during the experiment, generates and returns a summary message.
        The summary message is generated based on the most frequently occurring exception.

        Returns
        -------
        str
            Summary message for all aggregated errors,
        """
        logger.error("Several errors occurred during the experiment: %s", self.errors)

        errors_counter = Counter(error.__class__ for error in self.errors)
        most_common_error_type_name = errors_counter.most_common()[0][0].__name__

        error_content = next((er for er in self.errors if most_common_error_type_name in er.__class__.__name__))

        return f"{error_content}. " f"To find more details please see generated logs file."



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Any

from openai import OpenAI

from ai4rag.rag.embedding.base_model import BaseEmbeddingModel

# pylint: disable=duplicate-code


class OpenAIEmbeddingModel(BaseEmbeddingModel):
    """Class wrapping independently deployed embedding model with OpenAI client."""

    def __init__(self, client: OpenAI, model_id: str, params: dict[str, Any] | None = None):
        super().__init__(client=client, model_id=model_id, params=params)

    @property
    def params(self) -> dict[str, Any]:
        """Get model params."""
        return self._params

    @params.setter
    def params(self, params: dict[str, Any] | None) -> None:
        """Set model params."""
        if params is None:
            self._params = {}
        else:
            self._params = params
        if "embedding_dimension" not in self._params:
            self._params["embedding_dimension"] = self._detect_embedding_dimension()
        if "context_length" not in self._params:
            self._params["context_length"] = self._detect_context_length()

    def _detect_embedding_dimension(self) -> int:
        """Detect embedding dimension by making a minimal embedding call.

        Note: This method is called during initialization when ``embedding_dimension``
        is not present in the params dict.  It issues a real API request to the
        OpenAI API compliant endpoint, so the service must be reachable at construction time.

        Raises
        ------
        RuntimeError
            When the embedding dimension cannot be determined (e.g. the service
            is unreachable or the model is not available).
        """
        try:
            embedding = self.client.embeddings.create(model=self.model_id, input="test").data[0].embedding
        except Exception as exc:
            raise RuntimeError(
                f"Failed to auto-detect embedding dimension for model '{self.model_id}'. "
                "Provide 'embedding_dimension' explicitly or ensure the embedding service is reachable."
            ) from exc
        return len(embedding)

    def _detect_context_length(self) -> int:
        """Detect maximum context length via binary search over probe sizes.

        Performs a binary search between 64 and 8192 tokens to find the
        largest input the model accepts.  Each probe consists of repeated
        words so that one word ≈ one token.  The search stops when the
        remaining interval is smaller than 256 tokens, keeping the number
        of API calls to roughly 5.

        Raises
        ------
        RuntimeError
            When the context length cannot be determined (e.g. all probes fail).
        """
        lo, hi, best = 64, 8192, None
        while hi - lo >= 64:
            mid = (lo + hi) // 2
            probe_text = "word " * mid
            try:
                self.client.embeddings.create(model=self.model_id, input=probe_text)
                best = mid
                lo = mid + 1
            except Exception:  # pylint: disable=broad-exception-caught
                hi = mid - 1
        if best is not None:
            return best
        raise RuntimeError(
            f"Failed to auto-detect 'context_length' for model '{self.model_id}'. "
            "Provide 'context_length' explicitly or ensure the embedding service is reachable."
        )

    def embed_documents(self, texts: list[str]) -> list[list[float]]:
        """Embeds given list of strings.

        Parameters
        ----------
        texts : list[str]
            List of text-like chunks.

        Returns
        -------
        list[list[float]]
            Embeddings made from the list of texts.
        """
        resp = self.client.embeddings.create(model=self.model_id, input=texts)
        return [d.embedding for d in resp.data]

    def embed_query(self, query: str) -> list[float]:
        """Embeds given query.

        Parameters
        ----------
        query : str
            Single text-like chunk.

        Returns
        -------
        list[]
            Embeddings made from a single text.
        """
        return self.client.embeddings.create(model=self.model_id, input=query).data[0].embedding



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Annotated, Any

from annotated_types import Ge, Gt, Le
from openai import OpenAI
from pydantic import BaseModel

from ai4rag.rag.foundation_models.base_model import BaseFoundationModel, MessageTyped
from ai4rag.utils.constants import ChatGenerationConstants

# pylint: disable=duplicate-code


class OpenAIModelParameters(BaseModel):
    """Parameters to use for LSFoundationModel."""

    max_completion_tokens: Annotated[int, Gt(0)] = ChatGenerationConstants.MAX_COMPLETION_TOKENS
    temperature: Annotated[float, Ge(0), Le(1)] = ChatGenerationConstants.TEMPERATURE


class OpenAIFoundationModel(BaseFoundationModel[OpenAI, dict[str, Any] | OpenAIModelParameters | None]):
    """Wrapper for OpenAI client handled foundation models."""

    def __init__(
        self,
        client: OpenAI,
        model_id: str,
        params: dict[str, Any] | OpenAIModelParameters | None = None,
        system_message_text: str | None = None,
        user_message_text: str | None = None,
        context_template_text: str | None = None,
    ):
        super().__init__(
            client=client,
            model_id=model_id,
            params=params,
            system_message_text=system_message_text,
            user_message_text=user_message_text,
            context_template_text=context_template_text,
        )

    @property
    def params(self) -> OpenAIModelParameters:
        """Get model params."""
        return self._params

    @params.setter
    def params(self, params: dict | OpenAIModelParameters | None) -> None:
        """Set models params."""
        if isinstance(params, dict):
            self._params = OpenAIModelParameters(**params)
        elif isinstance(params, OpenAIModelParameters):
            self._params = params
        else:
            self._params = OpenAIModelParameters()

    def chat(self, messages: list[MessageTyped]) -> list[MessageTyped]:
        """
        Chat completion for communication with selected foundation model.

        Parameters
        ----------
        messages : list[MessageTyped]
            Messages to be included in the chat completion.

        Returns
        -------
        list[MessageTyped]
            Chat response choices from the model.
        """
        response_chat = self.client.chat.completions.create(
            model=self.model_id,
            messages=messages,
        )
        response_choices = response_chat.choices

        return response_choices



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from llama_stack_client import LlamaStackClient

from ..embedding.base_model import BaseEmbeddingModel
from .base_vector_store import BaseVectorStore
from .chroma import ChromaVectorStore
from .llama_stack import LSVectorStore


def get_vector_store(
    vs_type: str,
    embedding_model: BaseEmbeddingModel,
    reuse_collection_name: str | None = None,
    client: LlamaStackClient | None = None,
) -> BaseVectorStore:
    """Get vector store of desired type with chosen settings.

    Parameters
    ----------
    vs_type : str
        Type of vector store.

    embedding_model : BaseEmbeddingModel
        Embedding model used for the embeddings creation in the created vector store instance.

    reuse_collection_name : str | None, default=None
        Name of the collection that will be created in the vector database.

    client : LlamaStackClient | None, default=None
        Instance of the llama stack client to communicate with registered vector databases.

    Returns
    -------
    BaseVectorStore
        Instance of the vector store.
    """

    match vs_type:
        case "chroma":
            vs = ChromaVectorStore(
                embedding_model=embedding_model,
                reuse_collection_name=reuse_collection_name,
            )

        case "ls_milvus":
            vs = LSVectorStore(
                embedding_model=embedding_model,
                reuse_collection_name=reuse_collection_name,
                distance_metric="cosine",
                client=client,
                provider_id="milvus",
            )

        case _:
            raise ValueError(f"Vector store of type {vs_type} is not supported.")

    return vs
